/*
 *  RoombaComm Interface
 *
 *
 *
 * Copyright (c) 2005 Tod E. Kurt, tod@todbot.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA  02111-1307  USA
 *
 *
 *
 */

package roombacomm;


/**
 * RoombaComm -- the abstract base for all Roomba communications
 * 
 * This class contains the communications layer-independent parts of
 * how to communicate with a Roomba.  It does assume a very serial port-like
 * interaction, as the only working subclasses is for serial ports.
 * Thus, this objects form may change in the future.
 *
 * Standard lifecyle of this object (and its subclasses)
 *   RoombaComm roomba = new RoombaCommSubClass  (e.g. RoombaCommSerial)
 *   roomba.listports()
 *   roomba.connect("someportid");
 *   roomba.startup();
 *   while( ... ) {
 *      roomba.updateSensors();
 *      roomba.playNote( 53, 12 );
 *       
 *   roomba.disconnect();
 *
 *
 */
public abstract class RoombaComm
{

  boolean connected = false;
  boolean sensorsValid = false;
  boolean safetyFault = false;
  byte[] sensor_bytes = new byte[26];
  
  public boolean debug = false;

  public RoombaComm() {

  }

  /**
   * Return a list available portids, if applicable
   * or empty set if no ports, 
   * or return null if list is not enumerable
   */
  public abstract String[] listPorts();

  /**
   * returns true on successful connect, false otherwise
   * (for serial, portid is serial port name, for net, portid is url?)
   */
  public abstract  boolean connect(String portid);
  /**
   *
   */
  public abstract  void disconnect();


  /**
   * send given byte array to Roomba. 
   */
  public abstract boolean send(byte[] bytes);

  /** 
   * Send a single byte to the Roomba 
   * (defined as int because of stupid java signed bytes)
   */
  public abstract boolean send(int b);

  /**
   * returns true on successful sensor update, false otherwise
   * Subclasses should query Roomba and fill up 'sensor_bytes' with the full
   * sensor data set
   */
  public abstract boolean updateSensors();

  /**
   * Wake's Roomba up, if possible, thus optional
   */
  public void wakeup() {
    logmsg("subclass has not implemented");
  }

  /**
   * put Roomba in safe mode
   */
  public void startup() {
    logmsg("startup");
    byte cmd[] = { (byte)START, (byte)CONTROL };
    send( cmd );
  }

  /**
   *
   */
  public void powerOff() {
    logmsg("powerOff");
    //byte cmd[] = { (byte)POWER };
    send( (byte) POWER );
  }

  /** 
   * reset Roomba after a fault
   */
  public void reset() {
    logmsg("reset");
    startup();
    updateSensors();
  }

  //
  // basic functions
  //

  /** 
   * Just a simple pause function
   */
  public void pause( int millis ) {
    try { Thread.sleep(millis); } catch(Exception e) { }
  }

  /**
   * Plays a note via the hacky method of defining a one-note song & playing it
   * Uses up song slot 15
   */
  public void playNote( int note, int duration ) {
    logmsg("playnote: "+note+":"+duration);
    byte cmd[] = { (byte)SONG, 15, 1, (byte)note, (byte)duration,  // define song
                   (byte)PLAY, 15 };                               // play it bakc
    send( cmd );
  }

  /**
   * Move the Roomba via the low-level velocity + radius method
   */
  public void drive( int velocity, int radius ) {
    byte cmd[] = { (byte)DRIVE, (byte)(velocity >>> 8), (byte)(velocity & 0xff), 
                   (byte)(radius   >>> 8), (byte)(radius   & 0xff) };
    logmsg("drive: "+hex(cmd[0])+","+hex(cmd[1])+","+hex(cmd[2])+","+hex(cmd[3])+","+hex(cmd[4]));
    send( cmd );
  }

  /**
   * Turns on/off the non-drive motors (main brush, vacuum, sidebrush)
   */
  public void setMotors( boolean mainbrush, boolean vacuum, boolean sidebrush ) {
    byte cmd[] = { (byte)MOTORS, 
                   (byte)((mainbrush?0x04:0) | (vacuum?0x02:0) | (sidebrush?0x01:0)) };
    send( cmd );
  }

  /** 
   * turns on/off the various LEDs
   * FIXME: this is too complex
   */
  public void setLEDs( boolean status_green, boolean status_red, 
                       boolean spot, boolean clean, boolean max, boolean dirt, 
                       int power_color, int power_intensity ) {
    int v = (status_green?0x20:0) | (status_red?0x10:0) | (spot?0x08:0) | (clean?0x04:0) | (max?0x02:0) | (dirt?0x01:0);
    logmsg("setLEDS: "+binary(v));
    byte cmd[] = { (byte)LEDS, (byte)v, (byte)power_color, (byte)power_intensity };
    send(cmd);
  }


  //
  // higher-level functions
  //

  /**
   * Turn all vacuum motors on or off according to state
   */
  public void vacuum( boolean state ) {
    logmsg("vacuum: "+state);
    setMotors( state,state,state );
  }

  /**
   * Stop Rooomba's motion
   */
  public void stop() {
    logmsg("stop");
    drive( 0, 0 );
  }

  /**
   * Spin in place counter-clockwise, at specified speed
   */
  public void spinLeft( int speed ) {
    drive( speed, 1 ); 
  }

  /**
   * Spin in place clockwise, at specified speed
   */
  public void spinRight( int speed ) {
    drive( speed, -1 ); 
  }

  public void spinLeft() {
    spinLeft( 200 );  
  }
  public void spinRight() {
    spinRight( 200 );  
  }
  public void goStraight( int speed ) {
    drive( speed, 0x8000 );
  }
  public void goForward() {
     goStraight( 200 );
  }
  public void goBackward() {
     goStraight( -200 );
  }

  /**
   * if true, indicates we had an event that took the Roomba out of safe mode
   * called on every successful updateSensors
   * use the public boolean 'safety_fault' in your code
   */
  public boolean computeSafetyFault() {
    safetyFault = (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROP_MASK) != 0 ||
            sensor_bytes[CLIFFLEFT] == 1  || sensor_bytes[CLIFFFRONTLEFT] == 1 ||
            sensor_bytes[CLIFFRIGHT] == 1  || sensor_bytes[CLIFFFRONTRIGHT] == 1;
    return safetyFault;
  }

  public boolean bump() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMP_MASK) !=0;
  }  
  public boolean bumpLeft() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMPLEFT_MASK) !=0;
  }
  public boolean bumpRight() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMPRIGHT_MASK) !=0;
  }
  public boolean wheelDropLeft() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPLEFT_MASK) !=0;
  }
  public boolean wheelDropRight() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPRIGHT_MASK) !=0;
  }
  public boolean wheelDropCenter() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPCENT_MASK) !=0;
  }
  public boolean dirt() {
    int dl = sensor_bytes[DIRTLEFT];
    int dr = sensor_bytes[DIRTRIGHT];
    //if(debug) println("Roomba:dirt: dl,dr="+dl+","+dr);
    return (dl > 100) || (dr > 100);
  }

  public boolean cliffLeft() {
    return (sensor_bytes[CLIFFLEFT] != 0);
  }  
  public boolean cliffFrontLeft() {
    return (sensor_bytes[CLIFFFRONTLEFT] != 0);
  }  
  public boolean cliffFrontRight() {
    return (sensor_bytes[CLIFFFRONTRIGHT] != 0);
  }  
  public boolean cliffRight() {
    return sensor_bytes[CLIFFRIGHT] != 0;
  }
  
  public boolean powerButton() {
    return (sensor_bytes[BUTTONS] & POWERBUTTON_MASK) != 0;
  }
  public boolean spotButton() {
    return (sensor_bytes[BUTTONS] & SPOTBUTTON_MASK) != 0;
  }
  public boolean cleanButton() {
    return (sensor_bytes[BUTTONS] & CLEANBUTTON_MASK) != 0;
  }
  public boolean maxButton() {
    return (sensor_bytes[BUTTONS] & MAXBUTTON_MASK) != 0;
  }


  //
  // lower-level sensor access
  //
  public int bumps_wheeldrops() {
    return sensor_bytes[BUMPSWHEELDROPS];
  }
  public int wall() {
    return sensor_bytes[WALL];
  }
  public int cliff_left() {
    return sensor_bytes[CLIFFLEFT];
  }
  public int cliff_frontleft() {
    return sensor_bytes[CLIFFFRONTLEFT];
  }
  public int cliff_frontright() {
    return sensor_bytes[CLIFFFRONTRIGHT];
  }
  public int cliff_right() {
    return sensor_bytes[CLIFFRIGHT];
  }
  public int virtual_wall() {
    return sensor_bytes[VIRTUALWALL];
  }
  public int motor_overcurrents() {
    return sensor_bytes[MOTOROVERCURRENTS];
  }
  public int dirt_left() {
    return sensor_bytes[DIRTLEFT];
  }
  public int dirt_right() {
    return sensor_bytes[DIRTRIGHT];
  }

  public int remote_opcode() {
    return sensor_bytes[REMOTEOPCODE];
  }
  public int buttons() {
    return sensor_bytes[BUTTONS];
  }
  public int distance() {
    return sensor_bytes[DISTANCE_HI] << 8 | sensor_bytes[DISTANCE_LO];
  }
  public int angle() {
    return sensor_bytes[DISTANCE_HI] << 8 | sensor_bytes[ANGLE_LO];
  }
  
  public int charging_state() {
    return sensor_bytes[CHARGINGSTATE];
  }
  public int voltage() {
    return sensor_bytes[VOLTAGE_HI] << 8 | sensor_bytes[VOLTAGE_LO];
  }
  public int current() {
    return sensor_bytes[CURRENT_HI] << 8 | sensor_bytes[CURRENT_LO];
  }
  public int temperature() {
    return sensor_bytes[TEMPERATURE];
  }
  public int charge() {
    return sensor_bytes[CHARGE_HI] << 8 | sensor_bytes[CHARGE_LO];
  }
  public int capacity() {
    return sensor_bytes[CAPACITY_HI] << 8 | sensor_bytes[CAPACITY_LO];
  }


  // Roomba SCI opcodes
  // these should all be bytes, but Java bytes are signed, sucka
  static final int START   =  128;  // 0
  static final int BAUD    =  129;  // 1
  static final int CONTROL =  130;  // 0
  static final int SAFE    =  131;  // 0
  static final int FULL    =  132;  // 0
  static final int POWER   =  133;  // 0
  static final int SPOT    =  134;  // 0
  static final int CLEAN   =  135;  // 0
  static final int MAX     =  136;  // 0
  static final int DRIVE   =  137;  // 4
  static final int MOTORS  =  138;  // 1
  static final int LEDS    =  139;  // 3
  static final int SONG    =  140;  // 2N+2
  static final int PLAY    =  141;  // 1
  static final int SENSORS =  142;  // 1
  static final int DOCK    =  143;  // 0

  // offsets into sensor_bytes data
  static final int BUMPSWHEELDROPS     = 0;
  static final int WALL                = 1;
  static final int CLIFFLEFT           = 2;
  static final int CLIFFFRONTLEFT      = 3;
  static final int CLIFFFRONTRIGHT     = 4;
  static final int CLIFFRIGHT          = 5;
  static final int VIRTUALWALL         = 6;
  static final int MOTOROVERCURRENTS   = 7;
  static final int DIRTLEFT            = 8;
  static final int DIRTRIGHT           = 9;
  static final int REMOTEOPCODE        = 10;
  static final int BUTTONS             = 11;
  static final int DISTANCE_HI         = 12;
  static final int DISTANCE_LO         = 13;  
  static final int ANGLE_HI            = 14;
  static final int ANGLE_LO            = 15;
  static final int CHARGINGSTATE       = 16;
  static final int VOLTAGE_HI          = 17;
  static final int VOLTAGE_LO          = 18;  
  static final int CURRENT_HI          = 19;
  static final int CURRENT_LO          = 20;
  static final int TEMPERATURE         = 21;
  static final int CHARGE_HI           = 22;
  static final int CHARGE_LO           = 23;
  static final int CAPACITY_HI         = 24;
  static final int CAPACITY_LO         = 25;

  // bitmasks for various thingems
  static final int WHEELDROP_MASK      = 0x1C;
  static final int BUMP_MASK           = 0x03;
  static final int BUMPRIGHT_MASK      = 0x01;
  static final int BUMPLEFT_MASK       = 0x02;
  static final int WHEELDROPRIGHT_MASK = 0x04;
  static final int WHEELDROPLEFT_MASK  = 0x08;
  static final int WHEELDROPCENT_MASK  = 0x10;
  
  static final int POWERBUTTON_MASK    = 0x08;  
  static final int SPOTBUTTON_MASK     = 0x04;  
  static final int CLEANBUTTON_MASK    = 0x02;  
  static final int MAXBUTTON_MASK      = 0x01;  



  //
  // utility methods
  //

  //public void println(String s) {
  //  System.out.println(s);
  //}

  public String hex(int s) {
    return "not yet";
  }

  public String binary(int s) {
    return "not yet";
  }

  /**
   * just a little debug 
   */
  public void logmsg(String msg) {
    if(debug) 
      System.err.println("RoombaComm ("+System.currentTimeMillis()+"):"+msg);
  }

  /**
   * General error reporting, all corraled here just in case
   * I think of something slightly more intelligent to do.
   */
  public void errorMessage(String where, Throwable e) {
    e.printStackTrace();
    throw new RuntimeException("Error inside Serial." + where + "()");
  }

}

