/*
 *  RoombaComm Serial Interface
 *
 *  The serial-port based implementation of RoombaComm.
 *  Handles both physical RS-232 ports, USB adapter ports like Keyspan 
 *  USA-19HS, and Bluetooth serial port profiles.
 *
 * 
 * Some code taken from processing.serial.Serial.  Thanks guys!
 *
 *  Copyright (c) 2005 Tod E. Kurt, tod@todbot.com
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA  02111-1307  USA
 *
 */

package roombacomm;

import gnu.io.*;
import java.io.*;
import java.util.*;


public class RoombaCommSerial extends RoombaComm implements SerialPortEventListener
{
  public SerialPort port = null;
  static String    portname = null;   //"/dev/cu.KeySerial1";  // for instance
  static int           rate = 57600;
  static final int databits = 8;
  static final int parity   = SerialPort.PARITY_NONE;
  static final int stopbits = SerialPort.STOPBITS_1;

  static final int updateSensorsPause = 50;

  public InputStream input;
  public OutputStream output;

  byte buffer[] = new byte[32768];
  int bufferIndex;
  int bufferLast;

  int bufferSize = 1;  // how big before reset or event firing
  boolean bufferUntil;
  int bufferUntilByte;

  // constructor
  public RoombaCommSerial() {
    super();
  }

  /**
   *  
   */
  public boolean connect(String portid) {
    logmsg("connecting to port '"+portid+"'");
    portname = portid;

    connected = open_port();

    if( connected ) {
      wakeup();
      sensorsValid = false;
    }
    return connected;
  }

  public void disconnect() {
    try {
      // do io streams need to be closed first?
      if (input != null) input.close();
      if (output != null) output.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    input = null;
    output = null;
    
    try {
      if (port != null) port.close();  // close the port 
    } catch (Exception e) {
      e.printStackTrace();
    }
    port = null;
  }

  /**
   * subclassed.  FIXME: 
   */
  public boolean send(byte[] bytes) {
    try {
      output.write(bytes);
      output.flush();   // hmm, not sure if a good idea
    } catch (Exception e) { // null pointer or serial port dead
      //errorMessage("write", e);
      e.printStackTrace();
    }
    return true;
  }

  /**
   * This will handle both ints, bytes and chars transparently.
   */
  public boolean send(int b) {  // will also cover char
    try {
      output.write(b & 0xff);  // for good measure do the &
      output.flush();   // hmm, not sure if a good idea

    } catch (Exception e) { // null pointer or serial port dead
      errorMessage("write", e);
    }
    return true;
  }

  /**
   * toggles DD line via serial port DTR  (if available)
   */
  public void wakeup() {
    port.setDTR(false);
    try { Thread.sleep(500); } catch(Exception e) { }
    port.setDTR(true);
  }
  
  /**
   *
   */
  public boolean updateSensors() {
    byte cmd[] = { (byte)SENSORS, 0 };
    send(cmd);
    pause(updateSensorsPause);         // take a breather to let data come back
    sensorsValid = false;              // assume the worst, we're gothy
    int n = available();
    //logmsg("updateSensors:n="+n);
    if( n >= 26) {                     // there are enough bytes to read
      n = readBytes(sensor_bytes);
      if( n==26 ) {                    // did we get enough?
        sensorsValid = true;           // then everything's good, otherwise bad
        computeSafetyFault();
      }
    } else {
      logmsg("updateSensors:only "+n+" bytes available, not updating sensors");
    }
    //logmsg("buffer contains: "+ buffer );
    return sensorsValid;
  }

  /**
   * If this just hangs and never completes on Windows,
   * it may be because the DLL doesn't have its exec bit set.
   * Why the hell that'd be the case, who knows.
   * FIXME: deal more gracefully
   * (from processing.serial.Serial)
   */
  public String[] listPorts() {
    Vector list = new Vector();
    try {
      //System.err.println("trying");
      Enumeration portList = CommPortIdentifier.getPortIdentifiers();
      //System.err.println("got port list");
      while (portList.hasMoreElements()) {
        CommPortIdentifier portId =
          (CommPortIdentifier) portList.nextElement();
        //System.out.println(portId);

        if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
          String name = portId.getName();
          list.addElement(name);
        }
      }

    } catch (UnsatisfiedLinkError e) {
      //System.err.println("1");
      errorMessage("ports", e);

    } catch (Exception e) {
      //System.err.println("2");
      errorMessage("ports", e);
    }
    //System.err.println("move out");
    String outgoing[] = new String[list.size()];
    list.copyInto(outgoing);
    return outgoing;
  }


  // -------------------------------------------------------------
  // below only used internally to this class
  // -------------------------------------------------------------

  /**
   * internal method, used by connect()
   * FIXME: make it faile more gracefully, recognize bad port
   */
  private boolean open_port() {
    boolean success = false;
    try {
      Enumeration portList = CommPortIdentifier.getPortIdentifiers();
      while (portList.hasMoreElements()) {
        CommPortIdentifier portId =
          (CommPortIdentifier) portList.nextElement();
        
        if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
          //System.out.println("found " + portId.getName());
          if (portId.getName().equals(portname)) {
            logmsg("open_port:"+ portId);
            port = (SerialPort)portId.open("roomba serial madness", 2000);
            input  = port.getInputStream();
            output = port.getOutputStream();
            port.setSerialPortParams(rate, databits, stopbits, parity);
            port.addEventListener(this);
            port.notifyOnDataAvailable(true);
            logmsg("port "+portname+" opened successfully");
            success = true;
          }
        }
      }
      
    } catch (Exception e) {
      logmsg("connect failed: "+e);
      //errorMessage("<init>", e);
      //exception = e;
      //e.printStackTrace();
      port = null;
      input = null;
      output = null;
    }
    return success;
  }


  /**
   * callback for SerialPortEventListener
   * (from processing.serial.Serial)
   */
  synchronized public void serialEvent(SerialPortEvent serialEvent) {
    logmsg("serialEvent:"+serialEvent);
    if (serialEvent.getEventType() == SerialPortEvent.DATA_AVAILABLE) {
      try {
        while (input.available() > 0) {
          synchronized (buffer) {
            if (bufferLast == buffer.length) {
              byte temp[] = new byte[bufferLast << 1];
              System.arraycopy(buffer, 0, temp, 0, bufferLast);
              buffer = temp;
            }
            buffer[bufferLast++] = (byte) input.read();
            /*if (serialEventMethod != null) {
              if ((bufferUntil && (buffer[bufferLast-1] == bufferUntilByte)) ||
                  ((bufferLast - bufferIndex) >= bufferSize)) {
                try {
                  serialEventMethod.invoke(parent, new Object[] { this });
                } catch (Exception e) {
                  logmsg("error, disabling serialEvent() for " + port);
                  e.printStackTrace();
                  serialEventMethod = null;
                }
              }
            } */
          }
        }

      } catch (IOException e) {
        errorMessage("serialEvent", e);
      }
    }
  }

  /**
   * Returns the number of bytes that have been read from serial
   * and are waiting to be dealt with by the user.
   * (from processing.serial.Serial)
   */
  private int available() {
    return (bufferLast - bufferIndex);
  }

  /**
   * Return a byte array of anything that's in the serial buffer.
   * Not particularly memory/speed efficient, because it creates
   * a byte array on each read, but it's easier to use than
   * readBytes(byte b[]) (see below).
   * (from processing.serial.Serial)
   */
  private byte[] readBytes() {
    if (bufferIndex == bufferLast) return null;

    synchronized (buffer) {
      int length = bufferLast - bufferIndex;
      byte outgoing[] = new byte[length];
      System.arraycopy(buffer, bufferIndex, outgoing, 0, length);

      bufferIndex = 0;  // rewind
      bufferLast = 0;
      return outgoing;
    }
  }
  
  /**
   * Grab whatever is in the serial buffer, and stuff it into a
   * byte buffer passed in by the user. This is more memory/time
   * efficient than readBytes() returning a byte[] array.
   *
   * Returns an int for how many bytes were read. If more bytes
   * are available than can fit into the byte array, only those
   * that will fit are read.
   * (from processing.serial.Serial)
   */
  public int readBytes(byte outgoing[]) {
    if (bufferIndex == bufferLast) return 0;

    synchronized (buffer) {
      int length = bufferLast - bufferIndex;
      if (length > outgoing.length) length = outgoing.length;
      System.arraycopy(buffer, bufferIndex, outgoing, 0, length);

      bufferIndex += length;
      if (bufferIndex == bufferLast) {
        bufferIndex = 0;  // rewind
        bufferLast = 0;
      }
      return length;
    }
  }


}
